"""Auto-generated model profiles.

DO NOT EDIT THIS FILE MANUALLY.
This file is generated by the langchain-profiles CLI tool.

It contains data derived from the models.dev project.

Source: https://github.com/sst/models.dev
License: MIT License

To update these data, refer to the instructions here:

https://docs.langchain.com/oss/python/langchain/models#updating-or-overwriting-profile-data
"""

from typing import Any

_PROFILES: dict[str, dict[str, Any]] = {
    "cohere.command-r-plus-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-v2": {
        "max_input_tokens": 100000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-3-7-sonnet-20250219-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 8192,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-sonnet-4-20250514-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 64000,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "qwen.qwen3-coder-30b-a3b-v1:0": {
        "max_input_tokens": 262144,
        "max_output_tokens": 131072,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-2-11b-instruct-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 4096,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-3-haiku-20240307-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 4096,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-2-90b-instruct-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 4096,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-2-1b-instruct-v1:0": {
        "max_input_tokens": 131000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-v2:1": {
        "max_input_tokens": 200000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "deepseek.v3-v1:0": {
        "max_input_tokens": 163840,
        "max_output_tokens": 81920,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "cohere.command-light-text-v14": {
        "max_input_tokens": 4096,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "ai21.jamba-1-5-large-v1:0": {
        "max_input_tokens": 256000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-3-70b-instruct-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-3-opus-20240229-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 4096,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "amazon.nova-pro-v1:0": {
        "max_input_tokens": 300000,
        "max_output_tokens": 8192,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-1-8b-instruct-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "qwen.qwen3-32b-v1:0": {
        "max_input_tokens": 16384,
        "max_output_tokens": 16384,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-3-5-sonnet-20240620-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 8192,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-haiku-4-5-20251001-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 64000,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "cohere.command-r-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "amazon.nova-micro-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 8192,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-1-70b-instruct-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-70b-instruct-v1:0": {
        "max_input_tokens": 8192,
        "max_output_tokens": 2048,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "deepseek.r1-v1:0": {
        "max_input_tokens": 128000,
        "max_output_tokens": 32768,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-3-5-sonnet-20241022-v2:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 8192,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "cohere.command-text-v14": {
        "max_input_tokens": 4096,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-opus-4-20250514-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 32000,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "qwen.qwen3-coder-480b-a35b-v1:0": {
        "max_input_tokens": 131072,
        "max_output_tokens": 65536,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-sonnet-4-5-20250929-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 64000,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-2-3b-instruct-v1:0": {
        "max_input_tokens": 131000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-instant-v1": {
        "max_input_tokens": 100000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "amazon.nova-premier-v1:0": {
        "max_input_tokens": 1000000,
        "max_output_tokens": 16384,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-opus-4-1-20250805-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 32000,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama4-scout-17b-instruct-v1:0": {
        "max_input_tokens": 3500000,
        "max_output_tokens": 16384,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "ai21.jamba-1-5-mini-v1:0": {
        "max_input_tokens": 256000,
        "max_output_tokens": 4096,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama3-8b-instruct-v1:0": {
        "max_input_tokens": 8192,
        "max_output_tokens": 2048,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-3-sonnet-20240229-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 4096,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "meta.llama4-maverick-17b-instruct-v1:0": {
        "max_input_tokens": 1000000,
        "max_output_tokens": 16384,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "qwen.qwen3-235b-a22b-2507-v1:0": {
        "max_input_tokens": 262144,
        "max_output_tokens": 131072,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "amazon.nova-lite-v1:0": {
        "max_input_tokens": 300000,
        "max_output_tokens": 8192,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
    "anthropic.claude-3-5-haiku-20241022-v1:0": {
        "max_input_tokens": 200000,
        "max_output_tokens": 8192,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "pdf_tool_message": True,
        "image_tool_message": True,
    },
}
